unit BookMain;

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  StdCtrls, Buttons, ExtCtrls, ComCtrls,
  KeyU, VFileU, FFileU, IDGenU, SelectU;

type
  TfrmBook = class(TForm)
    labAuthor: TLabel;
    labTitle: TLabel;
    labPublisher: TLabel;
    labRemarks: TLabel;
    edtAuthor: TEdit;
    edtTitle: TEdit;
    edtPublisher: TEdit;
    memRemarks: TMemo;
    pnlNavigate: TPanel;
    gbxCategory: TGroupBox;
    lbxCat: TListBox;
    sbtAddCat: TSpeedButton;
    btnExit: TButton;
    btnCategories: TButton;
    pnlFind: TPanel;
    sbtFindTitle: TSpeedButton;
    sbtFindAuthor: TSpeedButton;
    sbtFindCat: TSpeedButton;
    pnlStore: TPanel;
    sbtNew: TSpeedButton;
    sbtClear: TSpeedButton;
    sbtFile: TSpeedButton;
    pnlNav: TPanel;
    sbtFirst: TSpeedButton;
    sbtPrev: TSpeedButton;
    sbtNext: TSpeedButton;
    sbtLast: TSpeedButton;
    pnlSpecial: TPanel;
    sbtDelete: TSpeedButton;
    sbtPrint: TSpeedButton;
    stbBook: TStatusBar;
    idgBooks: TIDGen;
    idgMemos: TIDGen;
    idgCatList: TIDGen;
    ffiBooks: TFFile;
    keyCat: TKey;
    keyTitle: TKey;
    keyAuthor: TKey;
    skyBooks: TSelKey;
    vfiMemos: TVFile;
    vfiCatList: TVFile;
    procedure btnExitClick(Sender: TObject);
    procedure FormCreate(Sender: TObject);
    procedure sbtClearClick(Sender: TObject);
    procedure sbtPrintClick(Sender: TObject);
    procedure sbtFindTitleClick(Sender: TObject);
    procedure sbtFindAuthorClick(Sender: TObject);
    procedure sbtFindCatClick(Sender: TObject);
    procedure sbtFirstClick(Sender: TObject);
    procedure sbtPrevClick(Sender: TObject);
    procedure sbtNextClick(Sender: TObject);
    procedure sbtLastClick(Sender: TObject);
    procedure sbtAddCatClick(Sender: TObject);
    procedure sbtNewClick(Sender: TObject);
    procedure sbtFileClick(Sender: TObject);
    procedure sbtDeleteClick(Sender: TObject);
    procedure FormCloseQuery(Sender: TObject; var CanClose: Boolean);
    procedure btnCategoriesClick(Sender: TObject);
  private
    { Private declarations }
    procedure DisplayHint(Sender: TObject);
  public
    { Public declarations }
  end;

var
  frmBook: TfrmBook;

type
  BookRec = record
    Author:    string[50];
    Title:     string[50];
    Publisher: string[40];
    Memo:      longint;
    Cats:      longint;
    end;

implementation

{$R *.DFM}

uses IDSetU, Work;

var
  ABook: BookRec;
  Mirrored: boolean;
  Altered: boolean;
  ChangedCats: boolean;
  Current: longint;
  Queued: word;
  Place: word;
  OrigCats:  TIDSet;
  RevisedCats: TIDSet;
  ShowSet: TIDSet;

{Helping procedures}

procedure ComputeAltered;

begin
  with frmBook do begin
    Altered := edtAuthor.Modified OR
               edtTitle.Modified OR
               edtPublisher.Modified OR
               memRemarks.Modified OR
               ChangedCats;
    end;
  end;


procedure GetIDsFromListBox;

var
  w: word;

begin
  with frmBook do begin
    RevisedCats.Empty;
    for w := 1 to lbxCat.Items.Count do
      RevisedCats.Add(keyCat.GetKeyID(lbxCat.Items[w-1]));
    end;
  end;


procedure SetAlteredFalse;

begin
  with frmBook do begin
    edtAuthor.Modified := FALSE;
    edtTitle.Modified := FALSE;
    edtPublisher.Modified := FALSE;
    memRemarks.Modified := FALSE;
    ChangedCats := FALSE;
    Altered := FALSE;
    end;
  end;


procedure BlankScreen;

begin
  with frmBook do begin
    edtAuthor.Text := '';
    edtTitle.Text := '';
    edtPublisher.Text := '';
    memRemarks.Text := '';
    lbxCat.Items.Clear;
    end;
  end;


function ValidBook(Code: integer): boolean;
{The Author and Title cannot be blank, and the Title cannot have
been used before. Code is 0 if we are inserting a new record and 1
if we are restoring an old one.}

var
  s: string;

begin
  Result := TRUE;  {Assume OK}
  s := '';
  with frmBook do begin
    if (edtAuthor.Text = '') or (edtAuthor.Text = ' ')
    then s := s + 'The Author cannot be blank.'+#10#13;
    if (edtTitle.Text = '') or (edtTitle.Text = ' ')
    then s := S + 'The Title cannot be blank.'+#10#13;
    if keyTitle.ExistKey(edtTitle.Text)
    then if Code = 0
         then s := s + 'The Title has been used before. '+#10#13
            + 'Titles must be unique.'
         else if edtTitle.Modified
              then s := s + 'The revised title is already in use. '+
                     #10#13+ 'Titles must be unique.';
    if s <> ''
    then begin
      ShowMessage(s);
      Result := FALSE;
      end;
    end;
  end;


procedure InsertBook;

var
  TL: word;
  w: word;

begin
  with ABook,frmBook do begin
    Author := edtAuthor.Text;
    Title := edtTitle.Text;
    Publisher := edtPublisher.Text;
    Current := idgBooks.Pop;
    if memRemarks.Modified
    then begin
      TL := memRemarks.GetTextLen+1;
      Memo := idgMemos.Pop;
      vfiMemos.Insert(Memo,pointer(memRemarks.Text),TL);
      end
    else Memo := -1;
    if ChangedCats
    then begin
      GetIDsFromListBox;
      if RevisedCats.Count > 0
      then begin
        Cats := idgCatList.Pop;
        vfiCatList.Insert(Cats,RevisedCats.SetPointer,
               RevisedCats.Count*4);
        for w := 1 to RevisedCats.Count do
          keyCat.IncludeRecID(lbxCat.Items[w-1],Current);
        end;
      end
    else Cats := -1;
    keyTitle.InsertKey(ABook.Title);
    keyTitle.IncludeRecID(ABook.Title,Current);
    if not keyAuthor.ExistKey(ABook.Author)
    then keyAuthor.InsertKey(ABook.Author);
    keyAuthor.IncludeRecID(ABook.Author,Current);
    ffiBooks.Insert(Current,@ABook);
    OrigCats.Empty;
    OrigCats.Append(RevisedCats);
    Mirrored := TRUE;
    SetAlteredFalse;
    end;
  end;


procedure RestoreBook;

var
  TL: word;
  w: word;
  WorkSet: TIDSet;

begin
  ComputeAltered;
  if not Altered
  then exit;
  with ABook,frmBook do begin
    {Did the user change Author?}
    if Author <> edtAuthor.Text
    then begin
      keyAuthor.RemoveRecID(Author,Current);
      keyAuthor.DeleteKey(Author,TRUE); {delete Author if no books}
      if not keyAuthor.ExistKey(edtAuthor.Text)
      then keyAuthor.InsertKey(edtAuthor.Text);
      keyAuthor.IncludeRecID(edtAuthor.Text,Current);
      end;
    Author := edtAuthor.Text;
    {Did the user change Title?}
    if Title <> edtTitle.Text
    then begin
      keyTitle.DeleteKey(Title,FALSE);
      keyTitle.InsertKey(edtTitle.Text);
      end;
    Title := edtTitle.Text;
    Publisher := edtPublisher.Text;
    {Did user change remarks?}
    if memRemarks.Modified
    then begin
      TL := memRemarks.GetTextLen+1;
      if TL = 1 {User deleted remarks}
      then begin
        vfiMemos.Delete(Memo);
        idgMemos.Push(Memo);
        Memo := -1;
        end
      else  {User changed remarks}
        if Memo = -1
        then begin
          Memo := idgMemos.Pop;
          vfiMemos.Insert(Memo,pointer(memRemarks.Text),TL);
          end
        else vfiMemos.Restore(Memo,pointer(memRemarks.Text),TL);
      end;
    {Did user change category list?}
    if ChangedCats
    then begin
      GetIDsFromListBox;
      if RevisedCats.Count = 0
      then begin {User deleted all categories}
        vfiCatList.Delete(Cats);
        idgCatList.Push(Cats);
        Cats := -1;
        end
      else vfiCatList.Restore(Cats,RevisedCats.SetPointer,
                RevisedCats.Count*4);
      {Now update the lists in the Category key}
      WorkSet := TIDSet.Create;
      WorkSet.Difference(RevisedCats,OrigCats);
      for w := 1 to WorkSet.Count do
        keyCat.IncludeRecID(keyCat.GetKey(WorkSet.Get(w-1)),Current);
      WorkSet.Difference(OrigCats,RevisedCats);
      for w := 1 to WorkSet.Count do
        keyCat.RemoveRecID(keyCat.GetKey(WorkSet.Get(w-1)),Current);
      WorkSet.Free;
      end;
    ffiBooks.Restore(Current,@ABook);
    OrigCats.Empty;
    OrigCats.Append(RevisedCats);
    Mirrored := TRUE;
    SetAlteredFalse;
    end;
  end;


procedure LoadBook;

var
  BL: word;
  w: word;
  MemoText: PChar;

begin
  with ABook,frmBook do begin
    Current := ShowSet.Get(Place);
    edtAuthor.Text := Author;
    edtTitle.Text := Title;
    edtPublisher.Text := Publisher;
    if Memo <> -1
    then begin
      BL := vfiMemos.GetBufSize(Memo);
      MemoText := StrAlloc(BL);
      vfiMemos.Fetch(Memo,pointer(MemoText));
      memRemarks.Text := MemoText;
      StrDispose(MemoText);
      end
    else memRemarks.Clear;
    lbxCat.Items.Clear;
    if Cats <> -1
    then begin
      BL := vfiCatList.GetBufSize(Cats);
      OrigCats.Capacity := BL div 4;
      OrigCats.Count := OrigCats.Capacity;
      vfiCatList.Fetch(Cats,OrigCats.SetPointer);
      for w := 1 to OrigCats.Count do
        lbxCat.Items.Add(keyCat.GetKey(OrigCats.Get(w-1)));
      end;
    end;
  Mirrored := TRUE;
  SetAlteredFalse;
  end;


procedure DeleteBook;

var
  w: word;

begin
  if Mirrored
  then with ABook,frmBook do begin
    if Memo <> -1
    then begin
      vfiMemos.Delete(Memo);
      idgMemos.Push(Memo);
      end;
    if Cats <> -1
    then begin
      vfiCatList.Delete(Cats);
      idgCatList.Push(Cats);
      for w := 1 to OrigCats.Count do
        keyCat.RemoveRecID(keyCat.GetKey(OrigCats.Get(w-1)),Current);
      end;
    keyAuthor.RemoveRecID(ABook.Author,Current);
    keyAuthor.DeleteKey(ABook.Author,TRUE);
    keyTitle.DeleteKey(ABook.Title,FALSE);
    ffiBooks.Delete(Current);
    idgBooks.Push(Current);
    end;
  OrigCats.Empty;
  Current := -1;
  BlankScreen;
  SetAlteredFalse;
  Mirrored := FALSE;
  end;


function OKToLoad: boolean;

begin
  Result := TRUE; {Assume OK}
  ComputeAltered;
  if Altered AND (frmBook.edtAuthor.Text <> '')
  then if Mirrored
       then if MessageDlg('Current record has been changed. '+
              'Should I restore it?',mtInformation,
              [mbYes,mbNo],0) = mrYes
            then if ValidBook(1)
                 then RestoreBook
                 else Result := FALSE
            else
       else if MessageDlg('Current record has not been added. '+
              'Should I add it?',mtInformation,
              [mbYes,mbNo],0) = mrYes
            then if ValidBook(0)
                 then InsertBook
                 else Result := FALSE
            else;
  end;



procedure LoadFirstRecord;

begin
  Queued := ShowSet.Count;
  if Queued > 1
  then frmBook.pnlNav.Color := clAqua
  else frmBook.pnlNav.Color := clBtnFace;
  Place := 0;
  if Queued <> 0
  then if OKToLoad
       then begin
         frmBook.ffiBooks.Fetch(ShowSet.Get(Place),@ABook);
         LoadBook;
         end;
  end;



procedure TfrmBook.DisplayHint(Sender: TObject);

begin
  stbBook.Panels[0].Text := Application.Hint;
  end;


procedure TfrmBook.FormCreate(Sender: TObject);
begin
  Application.OnHint := DisplayHint;
  idgBooks.AttachFile;
  idgMemos.AttachFile;
  idgCatList.AttachFile;
  ffiBooks.AttachFile;
  vfiMemos.AttachFile;
  vfiCatList.AttachFile;
  keyTitle.AttachFile;
  keyAuthor.AttachFile;
  keyCat.AttachFile;
  BlankScreen;
  SetAlteredFalse;
end;


procedure TfrmBook.sbtFindTitleClick(Sender: TObject);

var
  TitleList: TStringList;
  w: word;

begin
  TitleList := TStringList.Create;
  if skyBooks.SelectManyKeys(keyTitle,
          'Select books by title',TitleList)
  then begin
    ShowSet.Empty;
    for w := 1 to TitleList.Count do
      ShowSet.Add(keyTitle.GetRecID(TitleList[w-1],1));
    end;
  TitleList.Free;
  LoadFirstRecord;
  end;


procedure TfrmBook.sbtFindAuthorClick(Sender: TObject);

var
  AuthorList: TStringList;
  w,j: word;

begin
  AuthorList := TStringList.Create;
  if skyBooks.SelectManyKeys(keyAuthor,
          'Select books by author',AuthorList)
  then begin
    ShowSet.Empty;
    for w := 1 to AuthorList.Count do
      for j := 1 to keyAuthor.GetRecIDCount(AuthorList[w-1]) do
        ShowSet.Add(keyAuthor.GetRecID(AuthorList[w-1],j));
    end;
  AuthorList.Free;
  LoadFirstRecord;
  end;



procedure TfrmBook.sbtFindCatClick(Sender: TObject);

var
  CatList: TStringList;
  w,j: word;

begin
  CatList := TStringList.Create;
  if skyBooks.SelectManyKeys(keyCat,
          'Select books by category',CatList)
  then begin
    ShowSet.Empty;
    for w := 1 to CatList.Count do
      for j := 1 to keyCat.GetRecIDCount(CatList[w-1]) do
        ShowSet.Add(keyCat.GetRecID(CatList[w-1],j));
    end;
  CatList.Free;
  LoadFirstRecord;
  end;



procedure TfrmBook.sbtNewClick(Sender: TObject);
begin
  if ValidBook(0)
  then InsertBook;
end;



procedure TfrmBook.sbtClearClick(Sender: TObject);

begin
  if OKToLoad
  then begin
    BlankScreen;
    Mirrored := FALSE;
    SetAlteredFalse;
    Current := -1;
    ShowSet.Empty;
    pnlNav.Color := clBtnFace;
    ActiveControl := edtAuthor;
    end;
  end;


procedure TfrmBook.sbtFileClick(Sender: TObject);
begin
  if ValidBook(1)
  then RestoreBook;
end;



procedure TfrmBook.sbtFirstClick(Sender: TObject);
begin
  if Queued > 0
  then begin
    Place := 0;
    ffiBooks.Fetch(ShowSet.Get(Place),@ABook);
    if OKToLoad
    then LoadBook;
    end;
end;



procedure TfrmBook.sbtPrevClick(Sender: TObject);

begin
  if Queued > 0
  then if Place > 0
       then begin
         dec(Place);
         ffiBooks.Fetch(ShowSet.Get(Place),@ABook);
         if OKToLoad
         then LoadBook;
         end;
  end;



procedure TfrmBook.sbtNextClick(Sender: TObject);

begin
  if Queued > 0
  then if Place < Queued-1
       then begin
         inc(Place);
         ffiBooks.Fetch(ShowSet.Get(Place),@ABook);
         if OKToLoad
         then LoadBook;
         end;
  end;



procedure TfrmBook.sbtLastClick(Sender: TObject);

begin
  if Queued > 0
  then begin
    Place := Queued-1;
    ffiBooks.Fetch(ShowSet.Get(Place),@ABook);
    if OKToLoad
    then LoadBook;
    end;
  end;



procedure TfrmBook.sbtDeleteClick(Sender: TObject);
begin
  if not Mirrored
  then ShowMessage('Delete not needed. This record is not on file.')
  else if MessageDlg('Should I delete this record?',mtConFirmation,
           [mbYes,mbNo],0) = mrYes
       then DeleteBook;
end;




procedure TfrmBook.sbtPrintClick(Sender: TObject);
begin
  ShowMessage('Printing not implimented in this sample program.');
end;


procedure TfrmBook.btnExitClick(Sender: TObject);
begin
  Close;
end;


procedure TfrmBook.sbtAddCatClick(Sender: TObject);

var
  SelectedList: TStringList;
  InsertedList: TStringList;   {required but ignored}
  w: word;

begin
  SelectedList := TStringList.Create;
  InsertedList := TStringList.Create;
  for w := 1 to lbxCat.Items.Count do
    SelectedList.Add(lbxCat.Items[w-1]);
  if skyBooks.SelectManyKeysInsert(keyCat,
      'Categories for this book',InsertedList,SelectedList)
  then begin
    lbxCat.Items.Clear;
    for w := 1 to SelectedList.Count do
      lbxCat.Items.Add(SelectedList[w-1]);
    end;
  SelectedList.Free;
  InsertedList.Free;
  ChangedCats := TRUE;
end;


procedure TfrmBook.FormCloseQuery(Sender: TObject; var CanClose: Boolean);

begin
  OKToLoad;
  CanClose := TRUE;
end;


procedure TfrmBook.btnCategoriesClick(Sender: TObject);


begin
  ComputeAltered;
  if (Current <> -1) or Altered
  then ShowMessage(
     'Please finish editing your record and '+#10#13+
     'clear the screen before working with categories.')
  else frmWorkCat.ShowModal;
  end;


Initialization
  OrigCats := TIDSet.Create;
  RevisedCats := TIDSet.Create;
  ShowSet := TIDSet.Create;
  Current := -1;
  Queued := 0;
  Mirrored := FALSE;
finalization
  ShowSet.Free;
  RevisedCats.Free;
  OrigCats.Free;
end.


