unit IDGenU;

{$A-}  {[ack redords here}

interface

uses Classes;

type
  TIDGen = class(TComponent)
  private
    FFileName:       string;
    FIDFile:         file of longint;
    FAttached:       boolean;
    procedure  SetIDFileName(Value: string);
  published
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    procedure AttachFile;
    procedure DetachFile;
    function  Pop: longint;
    procedure Push(rid: longint);
    property  FileName: string
       read  FFileName
       write SetIDFileName;
    end;

{Property FileName: The name of a file that keeps track of id's.


 AttachFile: Hooks up the internal apparatus with the disk data
 set. The method MUST be called before any other methods will work.

 DetachFile: Breaks the connection between the TIDGen and the data
 set. Call this only if you wish to attach the TIDGen to another data
 set. For example

   AIDGen.DetachFile;
   AIDGen.IDFileName := 'NEWFILE';
   AIDGen.AttachFile;


 Pop: Returns an id for use. If any id's have been pushed onto the
 stack, one of them is poped and returned. If the stack is empty
 then a new never used before record id is returned.

 Push: Use this to add a record number to the recycle stack. When
 Pop next executes this record number will be returned. }



procedure Register;



implementation

uses Dialogs;


procedure Register;
begin
  RegisterComponents('Len''s',[TIDGen]);
  end;



{============ Setting properties =======================}

procedure TIDGen.SetIDFileName(Value: string);

var
  IDF: file of longint;

begin
  if Value <> FFileName
  then begin
    if Value = ''
    then if FAttached
         then DetachFile
         else
    else begin
      AssignFile(IDF,Value);
      {$I-} Reset(IDF); {$I+}
      if IOResult <> 0
      then begin
        ShowMessage('The file '+Value+' does not exist.');
        exit;
        end;
      CloseFile(IDF);
      end;
    FFileName := Value;
    end;
  end;




{=================  Methods of TIDGen ======================}

constructor TIDGen.Create(AOwner: TComponent);

begin
  inherited Create(AOwner);
  FFileName := '';
  FAttached := FALSE;
  end;


destructor TIDGen.Destroy;

begin
  DetachFile;
  inherited Destroy;
  end;



procedure TIDGen.AttachFile;

begin
  if FAttached
  then begin
    ShowMessage('File already attached');
    exit;
    end;
  FAttached := FALSE; {Assume the worst}
  if FFileName = ''
  then begin
    ShowMessage('No file named yet.');
    exit;
    end;
  AssignFile(FIDFile,FFileName);
  reset(FIDFile);
  FAttached := TRUE;
  end;


procedure TIDGen.DetachFile;

begin
  if not FAttached
  then exit;
  if FFileName <> ''
  then CloseFile(FIDFile);
  FFileName := '';
  FAttached := FALSE;
  end;



function TIDGen.Pop: longint;

var
  Next: longint;

begin
  if filesize(FIDFile) > 1
  then begin
    seek(FIDFile,filesize(FIDFile)-1);
    read(FIDFile,Result);
    seek(FIDFile,filesize(FIDFile)-1);
    Truncate(FIDFile);
    end
  else begin
    seek(FIDFile,0);
    read(FIDFile,Next);
    Result := Next;
    inc(Next);
    seek(FIDFile,0);
    write(FIDFile,Next);
    end;
  end;


procedure TIDGen.Push(rid: longint);

begin
  seek(FIDFile,filesize(FIDFile));
  write(FIDFile,rid);
  end;

end.
